# -*- coding: utf-8 -*-

## Geoid to Geoid (geoid2geoid) was developed for the purpose of transformation elevation models
## from one Geoid Model to another. These transformations are accomplished by using
## The Geoid models developed and maintained by NOAA's VDATUM transformation tool
## (https://vdatum.noaa.gov/). 

## !! Currently, the tool only provides methods for transforming a subset of GEOID models:
## GEOID99, GEOID03, GEOID09, GEOID12A, GEOID12B and GEOID18

## This tool is developed by the USGS Coastal National Elevation Database Program (CoNED)
## https://www.usgs.gov/coastal-changes-and-impacts/coned
## Written by Dean Tyler and Wm Matthew Cushing (mcushing@usgs.gov)
## This tool comes with no warranty

## This ArcGIS Pro Python Toolbox includes a Python toolbox script file *.pyt and a "geoidmodels"
## folder that includes the *.gtx and *.vrt files required to for GEOID transformations
## root --|
##        geoidmodels--|
##                     core--|
##                           Folders referencing each geoid model raster
##        geoid2geoid.pyt
##        geoid2geoid.pyt.xml

import arcpy
import os
import sys
import traceback

arcpy.CheckOutExtension('Spatial')
from arcpy import sa as sa

# Aliases
jn = os.path.join
splt = os.path.split
spltxt = os.path.splitext
env = arcpy.env
amsg = arcpy.AddMessage
wmsg = arcpy.AddWarning
emsg = arcpy.AddError

geoidsourcedict = {'Contiguous United States':'geoidmodels/core'}
geoidextlst = ['.gtx', '.tif', '.vrt']
geoiddict = {'GEOID99': ['g1999u0'], 'GEOID03': ['g2003u0'], 'GEOID09': ['g2009u0'],
            'GEOID12A': ['g2012au0'],'GEOID12B': ['g2012bu0'],'GEOID18': ['g2018u0']}

def errormsg():
    err_type = sys.exc_info()[0]
    err_value = sys.exc_info()[1]
    emsg(type(err_value))
    emsg(repr(err_value))
    tb = sys.exc_info()[2]
    tbinfo = traceback.format_tb(tb)[0]
    pymsg = f"PYTHON ERRORS:\nTraceback Info:\n {tbinfo}\n" \
            f"Error Info {str(err_type)}:{str(err_value)}\n\n"
    msgs = f"GP ERRORS:\n{arcpy.GetMessages(2)}\n\n"
    emsg(msgs)
    emsg(pymsg)

def getgeoidrst(geoid, corewksp):
    try:
        desc = arcpy.Describe(corewksp)
        wksptype = desc.workspaceType
        pth = desc.catalogPath
        for key, value in geoiddict.items():
            if key == geoid:
                if str(wksptype) == 'LocalDatabase':
                    rst = jn(pth,value[0])
                    if arcpy.Exists(rst):
                        return rst
                    else:
                        emsg(f'GEOID Raster does not exist: {rst}')
                        return None

                elif wksptype == 'FileSystem':
                    for ext in geoidextlst:
                        rst = jn(corewksp,key.lower(),f'{value[0]}{ext}')
                        if arcpy.Exists(rst):
                            return rst
    except:
        errormsg()
        return None

def geoid2geoid(inrst, corewksp, ingeoid, outgeoid, outrst, statstr='STATISTICS 1 1', cmprsstr='LZ77',
                pyrmdstr='PYRAMIDS -1 NEAREST LZ77 0 SKIP_FIRST', prllpf='', vb=False):
    try:
        desc = arcpy.Describe(inrst)
        extnt = desc.extent
        spat_ref = desc.spatialReference
        del desc

        outFolder = splt(outrst)[0]
        if vb: amsg(f'Output folder: {outFolder}')
        if vb: amsg(f'Input SR Name: {spat_ref.name}\tSR Factory Code: {spat_ref.factoryCode}\n')

        corefldr = jn(splt(os.path.abspath(__file__))[0], corewksp)
        ingeoidrst = getgeoidrst(ingeoid, corefldr)
        amsg(f'In GEOID: {ingeoidrst}')
        outgeoidrst = getgeoidrst(outgeoid, corefldr)
        amsg(f'Out GEOID: {outgeoidrst}')

        geoidsr = arcpy.Describe(ingeoidrst).spatialReference
        if vb: amsg(f'\nInput GEOID Model SR Name: {geoidsr.name}\tGEOID Model Factory Code: {geoidsr.factoryCode}')
        ### Clip in GEOID raster
        # Transform in Extent to in GEOID Extent
        if not spat_ref.type == geoidsr.type:
            if spat_ref.type.lower() == 'projected':
                extntbuf = 5000
            elif spat_ref.type.lower() == 'geographic':
                extntbuf = 0.045 #~5 km
            else:
                extntbuf = 0

            if vb: amsg(f'Input spatial extent: {extnt.XMin}, {extnt.YMin}, {extnt.XMax}, {extnt.YMax}')
            extntpnts = [arcpy.Point(extnt.XMin-extntbuf, extnt.YMin-extntbuf),
                        arcpy.Point(extnt.XMax+extntbuf, extnt.YMax+extntbuf)]
            extntgmtry = arcpy.Polyline(arcpy.Array(extntpnts), spat_ref)
            geoidextnt = extntgmtry.projectAs(geoidsr).extent
        else:
            amsg(f'Input Coordinate System same as GEOID model, no transformation required...\n')
            geoidextnt = extnt

        envparam ={'snapRaster':inrst, 'mask':inrst, 'scratchWorkspace':outFolder,
                    'compression':cmprsstr, 'rasterStatistics':statstr, 'resamplingMethod':'BILINEAR',
                    'pyramid':pyrmdstr, 'cellSize':'MINOF', 'overwriteOutput':True,
                    'parallelProcessingFactor':prllpf}

        with arcpy.EnvManager(workspace=outFolder, outputCoordinateSystem=geoidsr, extent = geoidextnt,
                                cellSize='MAXOF'):

            if vb: amsg(f'GEOID clipping Workspace: {env.workspace}\nGEOID clipping spatial extent: ' \
                        f'{geoidextnt.XMin} {geoidextnt.YMin} {geoidextnt.XMax} {geoidextnt.YMax}\n')

            try:
                wksptype = arcpy.Describe(env.workspace).workspaceType
                if wksptype == 'FileSystem':
                    ext = '.tif'
                else:
                    ext = ''
            except:
                emsg(f'Workspace Error')
                emsg(errormsg())
                return None

            if vb: amsg(f'Clipping current GEOID model: {ingeoidrst}')
            inclpgeoidfn = os.path.join(arcpy.env.workspace,f'ingeoidclp{ext}')
            inclpgeoid = sa.Times(ingeoidrst, 1)
            inclpgeoid.save(inclpgeoidfn)

            if vb: amsg(f'Clipping output GEOID model: {outgeoidrst}')
            outclpgeoidfn = os.path.join(arcpy.env.workspace,f'outgeoidclp{ext}')
            outclpgeoid = sa.Times(outgeoidrst, 1)
            outclpgeoid.save(outclpgeoidfn)
            if vb: amsg(f'Parallel Processing: {env.parallelProcessingFactor}\n')

            ingeoidrst = inclpgeoid
            outgeoidrst = outclpgeoid

        envparam['rasterStatistics'] = 'NONE'
        envparam['compression'] = 'NONE'
        envparam['pyramid'] = 'NONE'

        with arcpy.EnvManager(workspace=outFolder, outputCoordinateSystem=spat_ref, extent=extnt, **envparam):

            if vb: amsg(f'\nGEOID transformation Workspace: {env.workspace}\nGEOID transformation spatial extent: {env.extent.XMin}' \
                        f' {env.extent.YMin} {env.extent.XMax} {env.extent.YMax}')
            if vb: amsg(f'Parallel Processing: {env.parallelProcessingFactor}\n')
            ## Compress GEOID convertion calculation
            ## h = ellipsoidal height
            ## H = Terrain w/ GEOID Model
            ## N = GEOID Model
            ## h = H + N
            ########################## Converting Terrain GEIOD models equation ################################
            ## H(in) = Input DEM with existing GEOID Model, H(out) = Output DEM with desired GEOID model
            ## N(in) = Input GEOID Model [GEOID applied to H(in)], N(out) = Desired GEOID model converting to
            ## Map Algebra Function --> H(out) = [N(in) - N(out)]+ H(in)
            outdem = (sa.Raster(ingeoidrst) - sa.Raster(outgeoidrst)) + sa.Raster(inrst)
            ## example for finding ellipsoidal heights
            ## https://www.mathworks.com/help/map/ellipsoid-geoid-and-orthometric-height.html

        envparam['pyramid'] = pyrmdstr
        envparam['rasterStatistics'] = statstr
        envparam['compression'] = cmprsstr
        with arcpy.EnvManager(workspace=outFolder, outputCoordinateSystem=spat_ref, extent=extnt, **envparam):
            amsg(f'\nRaster compressing: {cmprsstr}\nBuild pyramids: {pyrmdstr}\nRaster Stats: {statstr} - ' \
                'Skip factor only applied with compression (idk)')
            if vb: amsg(f'Parallel Processing: {env.parallelProcessingFactor}')
            if not cmprsstr.lower() == 'none':
                arcpy.management.CopyRaster(outdem, outrst)
            else:
                outdem.save(outrst)

        clnuplst = [outclpgeoidfn, inclpgeoidfn]
        if vb: amsg('\nClean up...')
        for rst in clnuplst:
            if vb: amsg(f'\tDeleting {rst}')
            arcpy.management.Delete(rst)

        amsg('\nDone.\n')
        return outrst

    except:
        errormsg()
class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "Geoid2Geoid"
        self.alias = "Geoid-to-Geoid"
        self.description = "USGS CoNED Tool use to transform a raster from one GEOID to another GEOID"

        # List of tool classes associated with this toolbox
        self.tools = [geoid2geoidtool]

class geoid2geoidtool(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Geoid-to-Geoid"
        self.description = "USGS CoNED Tool use to transform a raster from one GEOID to another GEOID"
        self.canRunInBackground = True

    def getParameterInfo(self):
        """Define parameter definitions"""
        inraster = arcpy.Parameter(displayName='Input DEM dataset', name='inrst',
                                   datatype=['DERasterDataset', 'GPRasterLayer', 'GPMosaicLayer', 'DEMosaicDataset'],
                                   parameterType='Required', direction='Input', multiValue = False)

        ingeoid = arcpy.Parameter(displayName='Input DEM GEOID', name='ingeoid', parameterType='Required',
                                  direction='Input', datatype='GPString')

        outraster = arcpy.Parameter(displayName='Output DEM dataset', name='outrst',
                                    datatype=['DERasterDataset', 'GPRasterLayer'],
                                    parameterType='Required', direction='Output', multiValue = False)

        outgeoid = arcpy.Parameter(displayName='Output DEM GEOID', name='outgeoid', parameterType='Required',
                                   direction='Input', datatype='GPString')

        vregion = arcpy.Parameter(displayName='NOAA vertical datum region', name='vregion',
                                  datatype='GPString', parameterType='Required', direction='Input')

        prllpf = arcpy.Parameter(displayName='Parallel processing (n CPUs or n% CPUs)',
                                 name='prllpf', datatype='GPString', parameterType='Optional', direction='Input')

        statstr = arcpy.Parameter(displayName='Stat skip factor [n] \n(0 = No stats built)', name='stats',
                                  datatype='GPLong', parameterType='Required', direction='Input')

        cmprss = arcpy.Parameter(displayName='Compress dataset (LZ77) [~3x slower]', name='cmpressds', datatype='GPBoolean',
                                 parameterType='Required', direction='Input')

        bldpyrmds = arcpy.Parameter(displayName='Build pyramids [even slower]', name='bldpyrmds', datatype='GPBoolean',
                                    parameterType='Required', direction='Input')

        vb = arcpy.Parameter(displayName='Verbose (debugging)', name='vb', datatype='GPBoolean',
                                    parameterType='Required', direction='Input')

        vregion.filter.type = "ValueList"
        vregion.filter.list = list(geoidsourcedict.keys())
        vregion.value = vregion.filter.list[0]

        ingeoid.filter.type = "ValueList"
        ingeoid.filter.list = list(geoiddict.keys())
        ingeoid.value = ingeoid.filter.list[-2]

        outgeoid.filter.type = "ValueList"
        outgeoid.filter.list = list(geoiddict.keys())
        outgeoid.value = outgeoid.filter.list[-1]

        prllpf.value = 'Default'
        cmprss.value = True
        statstr.value = 1
        bldpyrmds.value = True
        vb.value = False
        params = [inraster, ingeoid, outraster, outgeoid, vregion, prllpf, statstr, cmprss, bldpyrmds, vb]
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        inrst = parameters[0].valueAsText
        ingeoid = parameters[1].value
        outrst = parameters[2].valueAsText
        outgeoid = parameters[3].value
        vregion = parameters[4].valueAsText
        prllpf = parameters[5].valueAsText
        statval = parameters[6].value
        cmprss = parameters[7].value
        bldpyrmds = parameters[8].value
        vb = parameters[9].value

        srcwksp = geoidsourcedict[vregion]

        if prllpf == None:
            prllpf = ''
        else:
            if prllpf.lower() == 'default' or prllpf.lower() == '':
                prllpf = ''
        # Build Stat String
        if statval == 0:
            statstr = 'NONE'
        else:
            statstr = f'STATISTICS {statval} {statval}'

        # Build Compression string
        if cmprss:
            cmprsstr = 'LZ77'
        else:
            cmprsstr = 'NONE'

        # Build Pyramids string
        if bldpyrmds:
            pyrmdstr = 'PYRAMIDS -1 NEAREST LZ77 0 SKIP_FIRST'
        else:
            pyrmdstr = 'NONE'
        amsg(f'Parallel Processing: {prllpf}')
        out = geoid2geoid(inrst, srcwksp, ingeoid, outgeoid, outrst, cmprsstr=cmprsstr, pyrmdstr=pyrmdstr, prllpf=prllpf,
                          statstr=statstr, vb=vb)
        return out

    def postExecute(self, parameters):
        """This method takes place after outputs are processed and
        added to the display."""
        return
